/*
 *  parser.c -- utilities used by the parser and lexical analyzer to
 *
 *      - read and unread characters from different input streams
 *	  (interactive or file);
 *
 *	- collect and retrieve errors
 */
#include "yacc2html.h"

#ifndef	MAX_ERRORS
#	define	MAX_ERRORS	15
#endif

typedef struct error {
	char  	     * message;
	struct error * next;
} ErrorRec, * Error;

extern int	  yylineno;			/* yacc file	*/

struct ParserInputRec parser_input;

static Error 	errors, * error_tail = &errors;
static int	n_errors;
static char	errbuf[1024];			/* XXX Unchecked. */

static int file_input()  {
	int 	c;

	if ((c = getc(parser_input.data.file.f)) == EOF) c = 0;
	else if (c == '\n') yylineno++;
	return c;
}

static int file_unput(c) int c; {
	if (c == '\n') yylineno--; 
	ungetc(c, parser_input.data.file.f);
	return c;
}

static int interactive_input()	{
	
	int 	c;

	if (parser_input.data.interactive.lookahead) {
		c = parser_input.data.interactive.lookahead;
		parser_input.data.interactive.lookahead = 0;
		if (c == '\n') yylineno++;
		return c;
	}
	if (parser_input.data.interactive.prompt) {
		parser_input.data.interactive.prompt = 0;

		fprintf(stderr, "> ");
		fflush(stderr);
	}

	if ((c = getc(parser_input.data.interactive.f)) == EOF) c = 0;
	else if (c == '\n') {
		parser_input.data.interactive.prompt = 1;
		yylineno++;
	}
	return c;
}

static int interactive_unput(c) int c;
{
	parser_input.data.interactive.lookahead = c;
	if (c == '\n') yylineno--;
	return c;
}

void parser_set_file(name, f) char * name; FILE * f;
{
	parser_input.type = INPUT_FILE;

	parser_input.data.file.f    = f;
	parser_input.data.file.name = name;

	parser_input.f_input = file_input;
	parser_input.f_unput = file_unput;
}

void parser_set_interactive(f) FILE * f;
{
	parser_input.type = INPUT_INTERACTIVE;

	parser_input.data.interactive.f = f;
	parser_input.data.interactive.lookahead = 0;
	parser_input.data.interactive.prompt = 1;

	parser_input.f_input = interactive_input;
	parser_input.f_unput = interactive_unput;
}

void parser_verror(fmt, ap) char const * fmt; va_list ap;
{
	if (n_errors > MAX_ERRORS) return;
	n_errors++;

	switch (parser_input.type) {
	case INPUT_INTERACTIVE:
		vfprintf(stderr, fmt, ap);
		fputc('\n', stderr);
		return;

	case INPUT_FILE:
		if (parser_input.data.file.name)
			sprintf(errbuf, "\"%s\", line %d: ",
			parser_input.data.file.name,  yylineno);
		else sprintf(errbuf, "standard input, line %d: ", yylineno);
		vsprintf(errbuf + strlen(errbuf), fmt, ap);
		break;

	default:
		fprintf(stderr, "parser_verror: unknown input type %d\n",
			parser_input.type);
		return;
	}

	*error_tail = TMALLOC( ErrorRec, 1);
	(*error_tail)->message = TMALLOC( char, strlen(errbuf) + 1);
	strcpy( (*error_tail)->message, errbuf );
	(*error_tail)->next = (Error)0;

	error_tail = &(*error_tail)->next;
}

#if defined(__STDC__) || defined(HAS_STDARG_H)

void parser_error(char const * fmt, ...)
{
	va_list	ap;
	va_start(ap, fmt);
	parser_verror(fmt, ap);
	va_end(ap);
}

#else

void parser_error(va_alist) va_dcl
{
	char * fmt;
	va_list ap;

	va_start(ap);
	fmt = va_arg(ap, char *);
	parser_verror(fmt, ap);
	va_end(ap);
}

#endif

int parser_print_errors()
{
	Error e, n;

	if (!errors) return 0;

	for (e = errors; e; e = n) {
		n = e->next;
		fputs(e->message, stderr);
		fputc('\n', stderr);
		VFREE(e);
	}
	errors = (Error)0;
	error_tail = &errors;

	return 1;
}
